// connect_sap.cpp
//
// A connection driver for SoftwareAuthority Protocol
//
//   (C) Copyright 2016 Fred Gleason <fredg@paravelsystems.com>
//
//   This program is free software; you can redistribute it and/or modify
//   it under the terms of the GNU General Public License as
//   published by the Free Software Foundation; either version 2 of
//   the License, or (at your option) any later version.
//
//   This program is distributed in the hope that it will be useful,
//   but WITHOUT ANY WARRANTY; without even the implied warranty of
//   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//   GNU General Public License for more details.
//
//   You should have received a copy of the GNU General Public
//   License along with this program; if not, write to the Free Software
//   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
//

#include <QStringList>

#include "connect_sap.h"

ConnectSap::ConnectSap(QObject *parent)
  : QObject(parent)
{
  sap_reading_sources=false;
  sap_reading_dests=false;

  //
  // The Socket
  //
  sap_socket=NULL;
  MakeSocket();

  //
  // Watchdog Timers
  //
  sap_holdoff_timer=new QTimer(this);
  sap_holdoff_timer->setSingleShot(true);
  connect(sap_holdoff_timer,SIGNAL(timeout()),
	  this,SLOT(holdoffReconnectData()));
}


ConnectSap::~ConnectSap()
{
}


unsigned ConnectSap::inputQuantity() const
{
  return sap_input_names.size();
}


QString ConnectSap::inputName(unsigned input)
{
  if(input>=sap_input_names.size()) {
    return QString();
  }
  return sap_input_names[input];
}


QString ConnectSap::inputLongName(unsigned input)
{
  if(input>=sap_input_long_names.size()) {
    return QString();
  }
  return sap_input_long_names[input];
}


unsigned ConnectSap::outputQuantity() const
{
  return sap_output_names.size();
}


QString ConnectSap::outputName(unsigned output)
{
  if(output>=sap_output_names.size()) {
    return QString();
  }
  return sap_output_names[output];
}


QString ConnectSap::outputLongName(unsigned output)
{
  if(output>=sap_output_long_names.size()) {
    return QString();
  }
  return sap_output_long_names[output];
}


unsigned ConnectSap::outputCrosspoint(unsigned output)
{
  return sap_output_xpoints[output];
}


void ConnectSap::setOutputCrosspoint(unsigned output,unsigned input)
{
  SendCommand(QString().sprintf("ActivateRoute %u %u %u",sap_matrix_number,
				output,input));
}


void ConnectSap::connectToHost(unsigned matrix_num,const QString &hostname,
			       uint16_t port,const QString &username,
			       const QString &passwd)
{
  QList<unsigned> empty;
  connectToHost(matrix_num,hostname,port,username,passwd,empty);
}


void ConnectSap::connectToHost(unsigned matrix_num,const QString &hostname,
			       uint16_t port,const QString &username,
			       const QString &passwd,
			       const QList<unsigned> &outputs)
{
  sap_matrix_number=matrix_num;
  sap_hostname=hostname;
  sap_port=port;
  sap_username=username;
  sap_password=passwd;
  sap_active_outputs=outputs;
  sap_socket->connectToHost(hostname,port);
}


void ConnectSap::connectedData()
{
  SendCommand("Login "+sap_username+" "+sap_password);
}


void ConnectSap::connectionClosedData()
{
  emit connected(false);
  sap_holdoff_timer->start(CONNECTSAP_HOLDOFF_INTERVAL);
}


void ConnectSap::holdoffReconnectData()
{
  MakeSocket();
  sap_socket->connectToHost(sap_hostname,sap_port);
}


void ConnectSap::readyReadData()
{
  QByteArray data;

  data=sap_socket->readAll();
  for(int i=0;i<data.length();i++) {
    switch(0xFFF&data[i]) {
    case 13:
      break;

    case 10:
      if(sap_accum.left(2)==">>") {
	sap_accum=sap_accum.right(sap_accum.length()-2);
      }
      DispatchCommand(sap_accum);
      sap_accum="";
      break;

    default:
      sap_accum+=data[i];
      break;
    }
  }
}


void ConnectSap::errorData(QAbstractSocket::SocketError err)
{
  sap_holdoff_timer->start(CONNECTSAP_HOLDOFF_INTERVAL);
}


void ConnectSap::DispatchCommand(const QString &cmd)
{
  //  printf("RECV: %s\n",(const char *)cmd.toUtf8());

  unsigned matrix;
  bool ok=false;

  QStringList f0=cmd.toLower().split(" ");

  //
  // Process Login
  //
  if(f0[0]=="login") {
    if((f0.size()==2)&&(f0[1]=="successful")) {
      SendCommand(QString().sprintf("SourceNames %u",sap_matrix_number));
    }
    else {
      emit connected(false);
    }
    return;
  }

  //
  // Check for delimiters
  //
  if(f0[0]=="begin") {
    if(f0.size()==4) {
      matrix=f0[3].toUInt(&ok);
      if(ok&&(matrix==sap_matrix_number)) {
	if(f0[1]=="sourcenames") {
	  sap_input_names.clear();
	  sap_input_long_names.clear();
	  sap_reading_sources=true;
	}
	if(f0[1]=="destnames") {
	  sap_output_names.clear();
	  sap_output_long_names.clear();
	  sap_reading_dests=true;
	}
      }
    }
    return;
  }
  if(f0[0]=="end") {
    if(f0.size()==4) {
      matrix=f0[3].toUInt(&ok);
      if(ok&&(matrix==sap_matrix_number)) {
	if(f0[1]=="sourcenames") {
	  sap_reading_sources=false;
	  emit inputListChanged();
	  SendCommand(QString().sprintf("DestNames %u",sap_matrix_number));
	}
	if(f0[1]=="destnames") {
	  emit outputListChanged();
	  if(sap_active_outputs.isEmpty()) {
	    SendCommand(QString().sprintf("RouteStat %u\r\n",
					  sap_matrix_number));
	  }
	  else {
	    for(int i=0;i<sap_active_outputs.size();i++) {
	    SendCommand(QString().sprintf("RouteStat %u %u\r\n",
					  sap_matrix_number,
					  sap_active_outputs[i]));
	    }
	  }
	  sap_reading_dests=false;
	  emit connected(true);
	}
      }
    }
    return;
  }

  //
  // Populate Endpoint Names
  //
  if(sap_reading_sources) {
    ReadSourceName(cmd);
    return;
  }
  if(sap_reading_dests) {
    ReadDestName(cmd);
    return;
  }

  //
  // Process Crosspoint Changes
  //
  if((f0[0]=="routestat")&&(f0.size()==5)) {
    matrix=f0[1].toUInt(&ok);
    if(ok&&(matrix==sap_matrix_number)) {
      unsigned output=f0[2].toUInt(&ok);
      if(ok) {
	unsigned input=f0[3].toUInt(&ok);
	if(ok) {
	  sap_output_xpoints[output]=input;
	  emit outputCrosspointChanged(output,input);
	}
      }
    }
  }

  //  printf("CMD: %s\n",(const char *)cmd.toUtf8());
}


void ConnectSap::ReadSourceName(const QString &cmd)
{
  QStringList f0=cmd.split("\t");

  if(f0.size()==8) {
    //
    // This blows up if the names are enumerated out of order!
    //
    sap_input_names.push_back(f0[1]);
    if(f0[6].toInt()<=0) {
      sap_input_long_names.push_back(f0[2]+" ["+tr("inactive")+"]");
    }
    else {
      sap_input_long_names.push_back(f0[2]+" ["+f0[6]+"]");
    }
  }
}


void ConnectSap::ReadDestName(const QString &cmd)
{
  QStringList f0=cmd.split("\t");

  if(f0.size()==6) {
    //
    // This blows up if the names are enumerated out of order!
    //
    sap_output_names.push_back(f0[1]);
    sap_output_long_names.push_back(f0[2]);
  }
}


void ConnectSap::BubbleSort(std::map<unsigned,QString> *names,
			    std::vector<unsigned> *ptrs)
{
  //
  // Reset Pointer Table
  //
  ptrs->clear();
  for(unsigned i=0;i<names->size();i++) {
    ptrs->push_back(i);
  }

  //
  // Sort
  //
  bool changed=true;
  while(changed) {
    changed=false;
    for(unsigned i=1;i<names->size();i++) {
      if((*names)[ptrs->at(i-1)]>(*names)[ptrs->at(i)]) {
	unsigned ptr=ptrs->at(i-1);
	ptrs->at(i-1)=ptrs->at(i);
	ptrs->at(i)=ptr;
	changed=true;
      }
    }
  }
}


void ConnectSap::SendCommand(const QString &cmd)
{
  //  printf("SendCommand(%s)\n",(const char *)cmd.toUtf8());
  sap_socket->write((cmd+"\r\n").toUtf8(),cmd.length()+2);
}


void ConnectSap::MakeSocket()
{
  if(sap_socket!=NULL) {
    delete sap_socket;
  }
  sap_socket=new QTcpSocket(this);
  connect(sap_socket,SIGNAL(connected()),this,SLOT(connectedData()));
  connect(sap_socket,SIGNAL(connectionClosed()),
	  this,SLOT(connectionClosedData()));
  connect(sap_socket,SIGNAL(readyRead()),this,SLOT(readyReadData()));
  connect(sap_socket,SIGNAL(error(QAbstractSocket::SocketError)),
	  this,SLOT(errorData(QAbstractSocket::SocketError)));
}
