// lwbuttonpanel.cpp
//
// Button applet for controlling an lwpath output.
//
//   (C) Copyright 2002-2016 Fred Gleason <fredg@paravelsystems.com>
//
//   This program is free software; you can redistribute it and/or modify
//   it under the terms of the GNU General Public License as
//   published by the Free Software Foundation; either version 2 of
//   the License, or (at your option) any later version.
//
//   This program is distributed in the hope that it will be useful,
//   but WITHOUT ANY WARRANTY; without even the implied warranty of
//   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//   GNU General Public License for more details.
//
//   You should have received a copy of the GNU General Public
//   License along with this program; if not, write to the Free Software
//   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
//

#include <QApplication>
#include <QMessageBox>

#include <sy/sycmdswitch.h>

#include "lwbuttonpanel.h"

//
// Icons
//
#include "../../icons/lwpath-16x16.xpm"

MainWidget::MainWidget(QWidget *parent)
  : QWidget(parent)
{
  panel_panel_number=0;

  QString conffile=PANEL_CONF_FILE;
  bool ok=false;

  //
  // Read Command Options
  //
  SyCmdSwitch *cmd=
    new SyCmdSwitch(qApp->argc(),qApp->argv(),"lwbuttonpanel",VERSION,
				   LWBUTTONPANEL_USAGE);
  for(unsigned i=0;i<cmd->keys();i++) {
    if(cmd->key(i)=="--configfile") {
      conffile=cmd->value(i);
      cmd->setProcessed(i,true);
    }
    if(cmd->key(i)=="--panel") {
      panel_panel_number=cmd->value(i).toUInt(&ok)-1;
      if(!ok) {
	QMessageBox::critical(this,"LWXYButtonPanel",
			      tr("Invalid --panel value!"));
	exit(256);
      }
      cmd->setProcessed(i,true);
    }
    if(!cmd->processed(i)) {
      QMessageBox::warning(this,"LWButtonPanel - "+tr("Error"),
			   tr("Unknown option")+": "+cmd->key(i)+"!");
      exit(256);
    }
  }

  //
  // Create And Set Icon
  //
  setWindowIcon(QPixmap(lwpath_16x16_xpm));

  //
  // Panel Configuration
  //
  panel_config=new PanelConfig();
  if(!panel_config->load(conffile)) {
    QMessageBox::warning(this,"LWButtonPanel - "+tr("Error"),
			 tr("Unable to open configuration file at")+
			 " \""+conffile+"\"!");
    exit(256);
  }
  if(panel_panel_number>=panel_config->panelQuantity()) {
    QMessageBox::critical(this,"LWXYButtonPanel",
			  tr("Non-existent panel specified!"));
    exit(256);
  }

  //
  // The SAP Connection
  //
  panel_connect=new ConnectSap(this);
  connect(panel_connect,SIGNAL(connected(bool)),
	  this,SLOT(changeConnectionState(bool)));
  connect(panel_connect,SIGNAL(inputListChanged()),
	  this,SLOT(updateInputNames()));
  connect(panel_connect,SIGNAL(outputListChanged()),
	  this,SLOT(updateOutputNames()));
  connect(panel_connect,
	  SIGNAL(outputCrosspointChanged(unsigned,unsigned)),
	  this,SLOT(changeOutputCrosspoint(unsigned,unsigned)));
  panel_button_mapper=new QSignalMapper(this);
  connect(panel_button_mapper,SIGNAL(mapped(int)),
	  this,SLOT(buttonClickedData(int)));

  //
  // Fire up the SAP connection
  //
  panel_connect->connectToHost(panel_config->matrixNumber(),
			       panel_config->serverHostname(),
			       panel_config->serverPort(),
			       panel_config->serverUsername(),
			       panel_config->serverPassword());
 
  delete cmd;
}


MainWidget::~MainWidget()
{
}


QSize MainWidget::sizeHint() const
{
  return QSize(10+100*panel_buttons.size(),70);
}


void MainWidget::buttonClickedData(int n)
{
  panel_connect->
    setOutputCrosspoint(panel_config->panelOutput(panel_panel_number),n+1);
}


void MainWidget::changeConnectionState(bool state)
{
  if(state) {
    for(unsigned i=0;i<panel_buttons.size();i++) {
      delete panel_buttons.at(i);
    }
    panel_buttons.clear();

    setWindowTitle(panel_connect->outputName(panel_panel_number)+" "+
		   tr("Switcher"));
    for(unsigned i=0;i<panel_connect->inputQuantity();i++) {
      panel_buttons.
	push_back(new QPushButton(panel_connect->inputName(i),this));
      panel_buttons.back()->show();
      connect(panel_buttons.back(),SIGNAL(clicked()),
	      panel_button_mapper,SLOT(map()));
      panel_button_mapper->setMapping(panel_buttons.back(),i);
    }
  }
  else {
    for(unsigned i=0;i<panel_buttons.size();i++) {
      panel_buttons.at(i)->setDisabled(true);
    }
  }

  resize(sizeHint());
}


void MainWidget::updateInputNames()
{
}


void MainWidget::updateOutputNames()
{
}


void MainWidget::changeOutputCrosspoint(unsigned output,unsigned input)
{
  if(output==panel_config->panelOutput(panel_panel_number)) {
    for(unsigned i=0;i<panel_connect->inputQuantity();i++) {
      if((input-1)==i) {
	panel_buttons.at(i)->setStyleSheet(LWPANELBUTTON_ACTIVE_STYLESHEET);
      }
      else {
	panel_buttons.at(i)->setStyleSheet("");
      }
    }
  }
}


void MainWidget::resizeEvent(QResizeEvent *e)
{
  for(unsigned i=0;i<panel_buttons.size();i++) {
    panel_buttons.at(i)->setGeometry(10+100*i,10,90,50);
  }
}


int main(int argc,char *argv[])
{
  QApplication a(argc,argv);

  MainWidget *w=new MainWidget();
  w->show();
  return a.exec();
}
