// livewire_map.h
//
// Marshalling routines for LiveWire maps
//
//   (C) Copyright 2015-2016 Fred Gleason <fredg@paravelsystems.com>
//
//   This program is free software; you can redistribute it and/or modify
//   it under the terms of the GNU General Public License version 2 as
//   published by the Free Software Foundation.
//
//   This program is distributed in the hope that it will be useful,
//   but WITHOUT ANY WARRANTY; without even the implied warranty of
//   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//   GNU General Public License for more details.
//
//   You should have received a copy of the GNU General Public
//   License along with this program; if not, write to the Free Software
//   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
//

#ifndef LIVEWIRE_MAP_H
#define LIVEWIRE_MAP_H

#include <stdint.h>

#include <map>
#include <vector>

#include <QHostAddress>
#include <QXmlDefaultHandler>

#include "livewire_audioport.h"
#include "livewire_gpio.h"
#include "livewire_node.h"

#define LIVEWIRE_DEFAULT_MAP_FILE "lwpathd_map.dat"

class LiveWireMapHandler : public QXmlDefaultHandler
{
 public:
  LiveWireMapHandler(std::vector<LiveWireNode *> *nodes,
		     std::vector<LiveWireAudioport *> *inputs,
		     std::vector<LiveWireAudioport *> *outputs,
		     std::vector<LiveWireGpio *> *gpis,
		     std::vector<LiveWireGpio *> *gpos);
  ~LiveWireMapHandler();
  bool startElement(const QString &namespaceURI,const QString &localName,
		    const QString &qName,const QXmlAttributes &atts);
  bool characters(const QString &ch);
  bool endElement(const QString &namespaceURI,const QString &localName,
		  const QString &qName);
  bool fatalError(const QXmlParseException &excp);
  QString errorString() const;

 private:
  enum Element {None=0,Node=1,Input=2,Output=3,Gpi=4,Gpo=5};
  bool ProcessNodeTag(const QString &tag,const QString &value);
  bool ProcessInputTag(const QString &tag,const QString &value);
  bool ProcessOutputTag(const QString &tag,const QString &value);
  bool ProcessGpiTag(const QString &tag,const QString &value);
  bool ProcessGpoTag(const QString &tag,const QString &value);
  std::vector<LiveWireNode *> *hand_nodes;
  std::vector<LiveWireAudioport *> *hand_inputs;
  std::vector<LiveWireAudioport *> *hand_outputs;
  std::vector<LiveWireGpio *> *hand_gpis;
  std::vector<LiveWireGpio *> *hand_gpos;
  Element hand_element;
  QString hand_accum;
  QString hand_error_string;
};




class LiveWireMap
{
 public:
  LiveWireMap();
  unsigned nodeQuantity() const;
  LiveWireNode *node(unsigned n);
  void addNode(const QHostAddress &addr,const QString &devname);
  unsigned inputQuantity() const;
  LiveWireAudioport *input(unsigned n);
  LiveWireAudioport *input(const QHostAddress &addr,unsigned slot);
  void addInput(unsigned number,const QHostAddress &addr,unsigned slot);
  unsigned outputQuantity() const;
  LiveWireAudioport *output(unsigned n);
  LiveWireAudioport *output(const QHostAddress &addr,unsigned slot);
  void addOutput(unsigned number,const QHostAddress &addr,unsigned slot);
  unsigned gpiQuantity() const;
  LiveWireGpio *gpi(unsigned n);
  LiveWireGpio *gpi(const QHostAddress &addr,unsigned slot);
  void addGpi(unsigned number,const QHostAddress &addr,unsigned slot);
  unsigned gpoQuantity() const;
  LiveWireGpio *gpo(unsigned n);
  LiveWireGpio *gpo(const QHostAddress &addr,unsigned slot);
  void addGpo(unsigned number,const QHostAddress &addr,unsigned slot);
  bool validate(QString *err_msgs);
  bool load(const QString &filename,QString *err_msg=NULL);
  bool save(const QString &filename) const;

 private:
  void IndexEndpoints();
  uint64_t SlotIndex(const QHostAddress &addr,unsigned slot) const;
  void WriteTag(FILE *f,const QString &tag,const QString &value,int indent=0)
    const;
  void WriteTag(FILE *f,const QString &tag,const QHostAddress &addr,
		int indent=0) const;
  void WriteTag(FILE *f,const QString &tag,unsigned value,int indent=0) const;
  QString EscapeXml(const QString &str) const;
  std::vector<LiveWireNode *> map_nodes;
  std::vector<LiveWireAudioport *> map_inputs;
  std::map<uint64_t,LiveWireAudioport *> map_input_index;
  std::vector<LiveWireAudioport *> map_outputs;
  std::map<uint64_t,LiveWireAudioport *> map_output_index;
  std::vector<LiveWireGpio *> map_gpis;
  std::map<uint64_t,LiveWireGpio *> map_gpi_index;
  std::vector<LiveWireGpio *> map_gpos;
  std::map<uint64_t,LiveWireGpio *> map_gpo_index;
};


#endif  // LIVEWIRE_MAP_H
