// lwpanel.cpp
//
// An applet for controling an LWPath output
//
//   (C) Copyright 2002-2016 Fred Gleason <fredg@paravelsystems.com>
//
//   This program is free software; you can redistribute it and/or modify
//   it under the terms of the GNU General Public License as
//   published by the Free Software Foundation; either version 2 of
//   the License, or (at your option) any later version.
//
//   This program is distributed in the hope that it will be useful,
//   but WITHOUT ANY WARRANTY; without even the implied warranty of
//   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//   GNU General Public License for more details.
//
//   You should have received a copy of the GNU General Public
//   License along with this program; if not, write to the Free Software
//   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
//

#include <stdlib.h>

#include <QApplication>
#include <QList>
#include <QMessageBox>
#include <QPainter>
#include <QTimer>
#include <QSettings>

#include <sy/sycmdswitch.h>

#include "lwpanel.h"

//
// Icons
//
#include "../../icons/lwpath-16x16.xpm"


MainWidget::MainWidget(QWidget *parent)
  :QWidget(parent)
{
  QString config_filename;

  //
  // Initialize Variables
  //
  QString hostname;
  unsigned panel_num;
  bool ok;
  unsigned cols=0;
  int ypos=0;
  panel_quantity=0;
  panel_columns=0;
  panel_rows=0;
  QList<unsigned> active_outputs;
  QString conffile=PANEL_CONF_FILE;
  setWindowTitle("LWPanel");

  //
  // Read Command Options
  //
  SyCmdSwitch *cmd=
    new SyCmdSwitch(qApp->argc(),qApp->argv(),"lwpanel",VERSION,LWPANEL_USAGE);
  for(unsigned i=0;i<cmd->keys();i++) {
    if(cmd->key(i)=="--columns") {
      panel_columns=cmd->value(i).toUInt();
      cmd->setProcessed(i,true);
    }
    if(cmd->key(i)=="--configfile") {
      conffile=cmd->value(i);
      cmd->setProcessed(i,true);
    }
    if(cmd->key(i)=="--panel") {  // Defer processing on these
      cmd->setProcessed(i,true);
    }
    if(!cmd->processed(i)) {
      QMessageBox::warning(this,"LWPanel - "+tr("Error"),
			   tr("Unknown option")+": "+cmd->key(i)+"!");
      exit(256);
    }
  }

  //
  // Create And Set Icon
  //
  setWindowIcon(QPixmap(lwpath_16x16_xpm));

  //
  // Panel Configuration
  //
  panel_config=new PanelConfig();
  if(!panel_config->load(conffile)) {
    QMessageBox::warning(this,"LWPanel - "+tr("Error"),
			 tr("Unable to open configuration file at")+
			 " \""+conffile+"\"!");
    exit(256);
  }

  //
  // The SAP Connection
  //
  panel_connect=new ConnectSap(this);

  //
  // The Button Clock
  //
  clock_state=false;
  QTimer *timer=new QTimer(this);
  timer->start(PANELWIDGET_FLASH_INTERVAL);
  this->connect(timer,SIGNAL(timeout()),this,SLOT(tickClock()));

  //
  // Start Output Panels
  //
  for(unsigned i=0;i<cmd->keys();i++) {
    if(cmd->key(i)=="--panel") {
      panel_num=cmd->value(i).toInt(&ok)-1;
      if(panel_num>=panel_config->panelQuantity()) {
	QMessageBox::warning(this,"LWPanel - "+tr("Error"),
			     tr("Panel")+QString().sprintf(" %u ",panel_num+1)+
			     tr("is undefined!"));
	exit(256);
      }
      if((ok)&&(panel_num<panel_config->panelQuantity())) {
	PanelWidget *widget=
	  new PanelWidget(panel_connect,panel_config,panel_num,this);
	connect(panel_connect,SIGNAL(connected(bool)),
		widget,SLOT(changeConnectionState(bool)));
	connect(panel_connect,SIGNAL(inputListChanged()),
		widget,SLOT(updateInputNames()));
	connect(panel_connect,SIGNAL(outputListChanged()),
		widget,SLOT(updateOutputNames()));
	connect(panel_connect,
		SIGNAL(outputCrosspointChanged(unsigned,unsigned)),
		widget,SLOT(changeOutputCrosspoint(unsigned,unsigned)));
	connect(this,SIGNAL(clockTicked(bool)),
		widget,SLOT(tickClock(bool)));
	if(panel_columns==0) {
	  widget->
	    setGeometry(10+panel_quantity*(20+widget->sizeHint().width()),0,
			widget->sizeHint().width(),
			widget->sizeHint().height());
	}
	else {
	  widget->
	    setGeometry(10+cols*(20+widget->sizeHint().width()),ypos,
			widget->sizeHint().width(),
			widget->sizeHint().height());
	  cols++;
	  if(cols==panel_columns) {
	    cols=0;
	    panel_rows++;
	    ypos+=(PANELWIDGET_HEIGHT+10);
	  }
	}
	active_outputs.push_back(panel_config->panelOutput(panel_num));
	panel_quantity++;
      }
    }
  }
  if(panel_quantity==0) {
    QMessageBox::warning(this,"LWPanel - "+tr("Error"),
			 tr("No panel specified!"));
    exit(256);
  }

  if(cols!=0) {
    panel_rows++;
  }
  /*
  if(panel_quantity==0) {
    QMessageBox::warning(this,"SASPanels",tr("No valid outputs to display!"));
    exit(256);
  }
  */

  //
  // Fix the Window Size
  //
  setMinimumWidth(sizeHint().width());
  setMaximumWidth(sizeHint().width());
  setMinimumHeight(sizeHint().height());
  setMaximumHeight(sizeHint().height());

  //
  // Fire up the SAP connection
  //
  panel_connect->connectToHost(panel_config->matrixNumber(),
			       panel_config->serverHostname(),
			       panel_config->serverPort(),
			       panel_config->serverUsername(),
			       panel_config->serverPassword(),
			       active_outputs);
 
  delete cmd;
}


MainWidget::~MainWidget()
{
}


QSize MainWidget::sizeHint() const
{
  if(panel_columns==0) {
    return QSize((PANELWIDGET_WIDTH+20)*panel_quantity,PANELWIDGET_HEIGHT+10);
  }
  else {
    return QSize((PANELWIDGET_WIDTH+20)*panel_columns,
		 (PANELWIDGET_HEIGHT+10)*panel_rows);
  }
}


void MainWidget::tickClock()
{
  emit clockTicked(clock_state);
  clock_state=!clock_state;
}


QSizePolicy MainWidget::sizePolicy() const
{
  return QSizePolicy(QSizePolicy::Fixed,QSizePolicy::Fixed);
}


void MainWidget::paintEvent(QPaintEvent *e)
{
  QPainter *p=new QPainter(this);
  p->setPen(Qt::black);
  p->setBrush(Qt::black);
  if(panel_columns==0) {
    for(unsigned i=1;i<panel_quantity;i++) {
      p->drawLine(i*(PANELWIDGET_WIDTH+20),0,
		  i*(PANELWIDGET_WIDTH+20),sizeHint().height());
    }
  }
  else {
    for(unsigned i=1;i<panel_columns;i++) {
      p->drawLine(i*(PANELWIDGET_WIDTH+20),0,
		  i*(PANELWIDGET_WIDTH+20),sizeHint().height());
    }
    for(unsigned i=1;i<panel_rows;i++) {
      p->drawLine(0,i*(PANELWIDGET_HEIGHT+10)-1,
		  sizeHint().width(),i*(PANELWIDGET_HEIGHT+10)-1);
    }
  }
  delete p;
}


int main(int argc,char *argv[])
{
  QApplication a(argc,argv);

  //
  // Start Event Loop
  //
  MainWidget *w=new MainWidget(NULL);
  //  a.setMainWidget(w);
  w->setGeometry(w->geometry().x(),w->geometry().y(),w->sizeHint().width(),w->sizeHint().height());
  w->show();
  return a.exec();
}
