// livewire_map.cpp
//
// Marshalling routines for LiveWire maps
//
//   (C) Copyright 2015-2016 Fred Gleason <fredg@paravelsystems.com>
//
//   This program is free software; you can redistribute it and/or modify
//   it under the terms of the GNU General Public License version 2 as
//   published by the Free Software Foundation.
//
//   This program is distributed in the hope that it will be useful,
//   but WITHOUT ANY WARRANTY; without even the implied warranty of
//   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//   GNU General Public License for more details.
//
//   You should have received a copy of the GNU General Public
//   License along with this program; if not, write to the Free Software
//   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
//

#include <sy/syprofile.h>

#include <QFile>
#include <QXmlInputSource>
#include <QXmlSimpleReader>

#include "livewire_map.h"

LiveWireMapHandler::LiveWireMapHandler(std::vector<LiveWireNode *> *nodes,
				       std::vector<LiveWireAudioport *> *inputs,
				       std::vector<LiveWireAudioport *> *outputs,
				       std::vector<LiveWireGpio *> *gpis,
				       std::vector<LiveWireGpio *> *gpos)
  : QXmlDefaultHandler()
{
  hand_nodes=nodes;
  hand_inputs=inputs;
  hand_outputs=outputs;
  hand_gpis=gpis;
  hand_gpos=gpos;
  hand_element=LiveWireMapHandler::None;
  hand_accum="";
  hand_error_string="OK";
}


LiveWireMapHandler::~LiveWireMapHandler()
{
}


bool LiveWireMapHandler::startElement(const QString &namespaceURI,
				      const QString &localName,
				      const QString &qName,
				      const QXmlAttributes &atts)
{
  if(localName=="Node") {
    hand_element=LiveWireMapHandler::Node;
    hand_nodes->push_back(new LiveWireNode());
  }
  if(localName=="Input") {
    hand_element=LiveWireMapHandler::Input;
    hand_inputs->push_back(new LiveWireAudioport());
  }
  if(localName=="Output") {
    hand_element=LiveWireMapHandler::Output;
    hand_outputs->push_back(new LiveWireAudioport());
  }
  if(localName=="Gpi") {
    hand_element=LiveWireMapHandler::Gpi;
    hand_gpis->push_back(new LiveWireGpio());
  }
  if(localName=="Gpo") {
    hand_element=LiveWireMapHandler::Gpo;
    hand_gpos->push_back(new LiveWireGpio());
  }
  hand_accum="";

  return true;
}


bool LiveWireMapHandler::characters(const QString &ch)
{
  hand_accum+=ch;
  return true;
}


bool LiveWireMapHandler::endElement(const QString &namespaceURI,
				    const QString &localName,
				    const QString &qName)
{
  bool ret=false;

  if((localName=="Node")||(localName=="Input")||(localName=="Output")||
     (localName=="Gpi")||(localName=="Gpo")||(localName=="LivewireMap")) {
    hand_element=LiveWireMapHandler::None;
    ret=true;
  }
  else {
    switch(hand_element) {
    case LiveWireMapHandler::Node:
      ret=ProcessNodeTag(localName,hand_accum);
      break;

    case LiveWireMapHandler::Input:
      ret=ProcessInputTag(localName,hand_accum);
      break;

    case LiveWireMapHandler::Output:
      ret=ProcessOutputTag(localName,hand_accum);
      break;

    case LiveWireMapHandler::Gpi:
      ret=ProcessGpiTag(localName,hand_accum);
      break;

    case LiveWireMapHandler::Gpo:
      ret=ProcessGpoTag(localName,hand_accum);
      break;

    case LiveWireMapHandler::None:
      break;
    }
  }

  if(!ret) {
    hand_error_string="ERROR: Tag: "+localName+"  Value: "+hand_accum;
  }

  return ret;
}


bool LiveWireMapHandler::fatalError(const QXmlParseException &excp)
{
  hand_error_string=QObject::tr("syntax error at line")+
    QString().sprintf(" %d",excp.lineNumber());
							    
  return false;
}


QString LiveWireMapHandler::errorString() const
{
  return hand_error_string;
}


bool LiveWireMapHandler::ProcessNodeTag(const QString &tag,const QString &value)
{
  if(tag=="Address") {
    if(!hand_nodes->back()->setHostAddress(value)) {
      return false;
    }
  }
  if(tag=="Password") {
    hand_nodes->back()->setPassword(value);
  }
  if(tag=="DeviceName") {
    hand_nodes->back()->setDeviceName(value);
  }
  return true;
}


bool LiveWireMapHandler::ProcessInputTag(const QString &tag,
					 const QString &value)
{
  unsigned num=0;
  bool ok=false;

  if(tag=="Number") {
    num=value.toUInt(&ok);
    if(!ok) {
      return false;
    }
    hand_inputs->back()->setNumber(num);
  }
  if(tag=="HostAddress") {
    if(!hand_inputs->back()->setHostAddress(value)) {
      if((!value.isEmpty())&&(value!="0.0.0.0")) {  // So we can load 'OFF' sources
	return false;
      }
    }
  }
  if(tag=="SlotNumber") {
    num=value.toUInt(&ok);
    if(!ok) {
      return false;
    }
    hand_inputs->back()->setSlotNumber(num-1);
  }
  return true;
}


bool LiveWireMapHandler::ProcessOutputTag(const QString &tag,
					  const QString &value)
{
  unsigned num=0;
  bool ok=false;

  if(tag=="Number") {
    num=value.toUInt(&ok);
    if(!ok) {
      return false;
    }
    hand_outputs->back()->setNumber(num);
  }
  if(tag=="HostAddress") {
    if(!hand_outputs->back()->setHostAddress(value)) {
      return false;
    }
  }
  if(tag=="SlotNumber") {
    num=value.toUInt(&ok);
    if(!ok) {
      return false;
    }
    hand_outputs->back()->setSlotNumber(num-1);
  }
  return true;
}


bool LiveWireMapHandler::ProcessGpiTag(const QString &tag,const QString &value)
{
  unsigned num=0;
  bool ok=false;

  if(tag=="Number") {
    num=value.toUInt(&ok);
    if(!ok) {
      return false;
    }
    hand_gpis->back()->setNumber(num);
  }
  if(tag=="HostAddress") {
    if(!hand_gpis->back()->setHostAddress(value)) {
      return false;
    }
  }
  if(tag=="SlotNumber") {
    num=value.toUInt(&ok);
    if(!ok) {
      return false;
    }
    hand_gpis->back()->setSlotNumber(num-1);
  }
  return true;
}


bool LiveWireMapHandler::ProcessGpoTag(const QString &tag,const QString &value)
{
  unsigned num=0;
  bool ok=false;

  if(tag=="Number") {
    num=value.toUInt(&ok);
    if(!ok) {
      return false;
    }
    hand_gpos->back()->setNumber(num);
  }
  if(tag=="HostAddress") {
    if(!hand_gpos->back()->setHostAddress(value)) {
      return false;
    }
  }
  if(tag=="SlotNumber") {
    num=value.toUInt(&ok);
    if(!ok) {
      return false;
    }
    hand_gpos->back()->setSlotNumber(num-1);
  }
  return true;
}




LiveWireMap::LiveWireMap()
{
}


unsigned LiveWireMap::nodeQuantity() const
{
  return map_nodes.size();
}


LiveWireNode *LiveWireMap::node(unsigned n)
{
  return map_nodes[n];
}


void LiveWireMap::addNode(const QHostAddress &addr,const QString &devname)
{
  map_nodes.push_back(new LiveWireNode(map_nodes.size()+1,addr,devname));
}


unsigned LiveWireMap::inputQuantity() const
{
  return map_inputs.size();
}


LiveWireAudioport *LiveWireMap::input(unsigned n)
{
  return map_inputs[n];
}


LiveWireAudioport *LiveWireMap::input(const QHostAddress &addr,unsigned slot)
{
  if(map_input_index.count(SlotIndex(addr,slot))==0) {
    return NULL;
  }
  return map_input_index[SlotIndex(addr,slot)];
}


void LiveWireMap::addInput(unsigned number,const QHostAddress &addr,
			   unsigned slot)
{
  map_inputs.push_back(new LiveWireAudioport(number,addr,slot));
  map_input_index[SlotIndex(addr,slot)]=map_inputs.back();
}


unsigned LiveWireMap::outputQuantity() const
{
  return map_outputs.size();
}


LiveWireAudioport *LiveWireMap::output(unsigned n)
{
  return map_outputs[n];
}


LiveWireAudioport *LiveWireMap::output(const QHostAddress &addr,unsigned slot)
{
  if(map_output_index.count(SlotIndex(addr,slot))==0) {
    return NULL;
  }
  return map_output_index[SlotIndex(addr,slot)];
}


void LiveWireMap::addOutput(unsigned number,const QHostAddress &addr,unsigned slot)
{
  map_outputs.push_back(new LiveWireAudioport(number,addr,slot));
  map_output_index[SlotIndex(addr,slot)]=map_outputs.back();
}


unsigned LiveWireMap::gpiQuantity() const
{
  return map_gpis.size();
}


LiveWireGpio *LiveWireMap::gpi(unsigned n)
{
  return map_gpis[n];
}


LiveWireGpio *LiveWireMap::gpi(const QHostAddress &addr,unsigned slot)
{
  if(map_gpi_index.count(SlotIndex(addr,slot))==0) {
    return NULL;
  }
  return map_gpi_index[SlotIndex(addr,slot)];
}


void LiveWireMap::addGpi(unsigned number,const QHostAddress &addr,unsigned slot)
{
  map_gpis.push_back(new LiveWireGpio(number,addr,slot));
  map_gpi_index[SlotIndex(addr,slot)]=map_gpis.back();
}


unsigned LiveWireMap::gpoQuantity() const
{
  return map_gpos.size();
}


LiveWireGpio *LiveWireMap::gpo(unsigned n)
{
  return map_gpos[n];
}


LiveWireGpio *LiveWireMap::gpo(const QHostAddress &addr,unsigned slot)
{
  if(map_gpo_index.count(SlotIndex(addr,slot))==0) {
    return NULL;
  }
  return map_gpo_index[SlotIndex(addr,slot)];
}


void LiveWireMap::addGpo(unsigned number,const QHostAddress &addr,unsigned slot)
{
  map_gpos.push_back(new LiveWireGpio(number,addr,slot));
  map_gpo_index[SlotIndex(addr,slot)]=map_gpos.back();
}


bool LiveWireMap::validate(QString *err_msgs)
{
  unsigned errs=0;

  //
  // Check for Orphaned/Duplicate Node Host Addresses
  //
  for(unsigned i=0;i<map_nodes.size();i++) {
    for(unsigned j=i;j<map_nodes.size();j++) {
      if((i!=j)&&(map_nodes[i]->hostAddress()==map_nodes[j]->hostAddress())) {
	*err_msgs+=QObject::tr("address")+" "+
	  map_nodes[i]->hostAddress().toString()+" "+
	  QObject::tr("is assigned to more than one node.")+"\n";
	errs++;
      }
    }
  }

  //
  // Check for Orphaned/Missing/Duplicate Output Numbers
  //
  for(unsigned i=0;i<map_outputs.size();i++) {
    bool found=false;
    for(unsigned j=0;j<map_nodes.size();j++) {
      if(map_outputs[i]->hostAddress()==map_nodes[j]->hostAddress()) {
	found=true;
      }
    }
    if((!found)&&(map_outputs[i]->hostAddress().toIPv4Address()!=0)) {
      *err_msgs+=QObject::tr("output")+QString().sprintf(" %u ",i+1)+
	QObject::tr("has no parent node.")+"\n";
      errs++;
    }
    if(map_outputs[i]->number()>map_outputs.size()) {
      *err_msgs+=QObject::tr("output number")+
	QString().sprintf(" %u ",map_outputs[i]->number())+
	QObject::tr("is outside of legal range.")+"\n";
      errs++;
    }
    else {
      for(unsigned j=i;j<map_outputs.size();j++) {
	if((i!=j)&&(map_outputs[i]->number()==map_outputs[j]->number())) {
	  *err_msgs+=QObject::tr("output number")+
	    QString().sprintf(" %u ",map_outputs[i]->number())+
	    QObject::tr("is duplicated.")+"\n";
	  errs++;
	}
      }
    }
  }

  //
  // Check for Orphaned/Missing/Duplicate Input Numbers
  //
  for(unsigned i=0;i<map_inputs.size();i++) {
    bool found=false;
    for(unsigned j=0;j<map_nodes.size();j++) {
      if((map_inputs[i]->hostAddress()==map_nodes[j]->hostAddress())||
	 (map_inputs[i]->hostAddress().isNull())) {
	found=true;
      }
    }
    if((!found)&&(map_inputs[i]->hostAddress().toIPv4Address()!=0)) {
      *err_msgs+=QObject::tr("input")+QString().sprintf(" %u ",i+1)+
	QObject::tr("has no parent node.")+"\n";
      errs++;
    }
    if(map_inputs[i]->number()>map_inputs.size()) {
      *err_msgs+=QObject::tr("input number")+
	QString().sprintf(" %u ",map_inputs[i]->number())+
	QObject::tr("is outside of legal range.")+"\n";
      errs++;
    }
    else {
      for(unsigned j=i;j<map_inputs.size();j++) {
	if((i!=j)&&(map_inputs[i]->number()==map_inputs[j]->number())) {
	  *err_msgs+=QObject::tr("input number")+
	    QString().sprintf(" %u ",map_inputs[i]->number())+
	    QObject::tr("is duplicated.")+"\n";
	  errs++;
	}
      }
    }
  }

  //
  // Check for Orphaned/Missing/Duplicate GPI Numbers
  //
  for(unsigned i=0;i<map_gpis.size();i++) {
    bool found=false;
    for(unsigned j=0;j<map_nodes.size();j++) {
      if(map_gpis[i]->hostAddress()==map_nodes[j]->hostAddress()) {
	found=true;
      }
    }
    if((!found)&&(map_gpis[i]->hostAddress().toIPv4Address()!=0)) {
      *err_msgs+=QObject::tr("gpi")+QString().sprintf(" %u ",i+1)+
	QObject::tr("has no parent node.")+"\n";
      errs++;
    }
    if(map_gpis[i]->number()>map_gpis.size()) {
      *err_msgs+=QObject::tr("gpi number")+
	QString().sprintf(" %u ",map_gpis[i]->number())+
	QObject::tr("is outside of legal range.")+"\n";
      errs++;
    }
    else {
      for(unsigned j=i;j<map_gpis.size();j++) {
	if((i!=j)&&(map_gpis[i]->number()==map_gpis[j]->number())) {
	  *err_msgs+=QObject::tr("gpi number")+
	    QString().sprintf(" %u ",map_gpis[i]->number())+
	    QObject::tr("is duplicated.")+"\n";
	  errs++;
	}
      }
    }
  }

  //
  // Check for Missing/Duplicate GPO Numbers
  //
  for(unsigned i=0;i<map_gpos.size();i++) {
    bool found=false;
    for(unsigned j=0;j<map_nodes.size();j++) {
      if(map_gpos[i]->hostAddress()==map_nodes[j]->hostAddress()) {
	found=true;
      }
    }
    if((!found)&&(map_gpos[i]->hostAddress().toIPv4Address()!=0)) {
      *err_msgs+=QObject::tr("gpo")+QString().sprintf(" %u ",i+1)+
	QObject::tr("has no parent node.")+"\n";
      errs++;
    }


    if(map_gpos[i]->number()>map_gpos.size()) {
      *err_msgs+=QObject::tr("gpo number")+
	QString().sprintf(" %u ",map_gpos[i]->number())+
	QObject::tr("is outside of legal range.")+"\n";
      errs++;
    }
    else {
      for(unsigned j=i;j<map_gpos.size();j++) {
	if((i!=j)&&(map_gpos[i]->number()==map_gpos[j]->number())) {
	  *err_msgs+=QObject::tr("gpo number")+
	    QString().sprintf(" %u ",map_gpos[i]->number())+
	    QObject::tr("is duplicated.")+"\n";
	  errs++;
	}
      }
    }
  }

  return errs==0;
}


bool LiveWireMap::load(const QString &filename,QString *err_msg)
{
  bool ret=false;
  QXmlSimpleReader *reader=new QXmlSimpleReader();
  QFile *file=new QFile(filename);
  if(!file->open(QIODevice::ReadOnly)) {
    if(err_msg!=NULL) {
      *err_msg=QObject::tr("unable to open map file");
    }
    return false;
  }
  QXmlInputSource *xmlsrc=new QXmlInputSource(file);
  LiveWireMapHandler *handler=
    new LiveWireMapHandler(&map_nodes,&map_inputs,&map_outputs,
			   &map_gpis,&map_gpos);
  reader->setContentHandler(handler);
  reader->setErrorHandler(handler);

  ret=reader->parse(xmlsrc);
  if((!ret)&&(err_msg!=NULL)) {
    *err_msg=handler->errorString();
  }
  IndexEndpoints();

  delete handler;
  delete xmlsrc;
  delete file;
  delete reader;

  return ret;
}


bool LiveWireMap::save(const QString &filename) const
{
  FILE *f=NULL;

  if(filename.isEmpty()) {
    f=stdout;
  }
  else {
    f=fopen((filename+".temp").toUtf8(),"w");
  }
  if(f!=NULL) {
    fprintf(f,"<LivewireMap>\n");

    //
    // Save Node Map
    //
    for(unsigned i=0;i<map_nodes.size();i++) {
      fprintf(f,"  <Node>\n");
      WriteTag(f,"Address",map_nodes[i]->hostAddress(),4);
      WriteTag(f,"Password",map_nodes[i]->password(),4);
      WriteTag(f,"DeviceName",map_nodes[i]->deviceName(),4);
      fprintf(f,"  </Node>\n");
    }

    //
    // Save Output Map
    //
    for(unsigned i=0;i<map_outputs.size();i++) {
      fprintf(f,"  <Output>\n");
      WriteTag(f,"Number",map_outputs[i]->number(),4);
      WriteTag(f,"HostAddress",map_outputs[i]->hostAddress(),4);
      WriteTag(f,"SlotNumber",map_outputs[i]->slotNumber()+1,4);
      fprintf(f,"  </Output>\n");
    }

    //
    // Save Input Map
    //
    for(unsigned i=0;i<map_inputs.size();i++) {
      fprintf(f,"  <Input>\n");
      WriteTag(f,"Number",map_inputs[i]->number(),4);
      WriteTag(f,"HostAddress",map_inputs[i]->hostAddress(),4);
      WriteTag(f,"SlotNumber",map_inputs[i]->slotNumber()+1,4);
      fprintf(f,"  </Input>\n");
    }

    //
    // Save GPI Map
    //
    for(unsigned i=0;i<map_gpis.size();i++) {
      fprintf(f,"  <Gpi>\n");
      WriteTag(f,"Number",map_gpis[i]->number(),4);
      WriteTag(f,"HostAddress",map_gpis[i]->hostAddress(),4);
      WriteTag(f,"SlotNumber",map_gpis[i]->slotNumber()+1,4);
      fprintf(f,"  </Gpi>\n");
    }

    //
    // Save GPO Map
    //
    for(unsigned i=0;i<map_gpos.size();i++) {
      fprintf(f,"  <Gpo>\n");
      WriteTag(f,"Number",map_gpos[i]->number(),4);
      WriteTag(f,"HostAddress",map_gpos[i]->hostAddress(),4);
      WriteTag(f,"SlotNumber",map_gpos[i]->slotNumber()+1,4);
      fprintf(f,"  </Gpo>\n");
    }

    fprintf(f,"</LivewireMap>\n");

    fclose(f);
    rename((filename+".temp").toUtf8(),filename.toUtf8());
  }
  else {
    return false;
  }
  return true;
}


void LiveWireMap::IndexEndpoints()
{
  map_input_index.clear();
  for(unsigned i=0;i<map_inputs.size();i++) {
    map_input_index[SlotIndex(map_inputs[i]->hostAddress(),map_inputs[i]->slotNumber())]=map_inputs[i];
  }

  map_output_index.clear();
  for(unsigned i=0;i<map_outputs.size();i++) {
    map_output_index[SlotIndex(map_outputs[i]->hostAddress(),map_outputs[i]->slotNumber())]=map_outputs[i];
  }

  map_gpi_index.clear();
  for(unsigned i=0;i<map_gpis.size();i++) {
    map_gpi_index[SlotIndex(map_gpis[i]->hostAddress(),map_gpis[i]->slotNumber())]=map_gpis[i];
  }

  map_gpo_index.clear();
  for(unsigned i=0;i<map_gpos.size();i++) {
    map_gpo_index[SlotIndex(map_gpos[i]->hostAddress(),map_gpos[i]->slotNumber())]=map_gpos[i];
  }
}


uint64_t LiveWireMap::SlotIndex(const QHostAddress &addr,unsigned slot) const
{
  return (uint64_t)addr.toIPv4Address()+(((uint64_t)slot)<<32);
}


void LiveWireMap::WriteTag(FILE *f,const QString &tag,const QString &value,
			   int indent) const
{
  for(int i=0;i<indent;i++) {
    fprintf(f," ");
  }
  if(value.isEmpty()) {
    fprintf(f,"<%s/>\n",(const char *)tag.toUtf8());
  }
  else {
    fprintf(f,"<%s>%s</%s>\n",(const char *)tag.toUtf8(),
	    (const char *)EscapeXml(value).toUtf8(),(const char *)tag.toUtf8());
  }
}


void LiveWireMap::WriteTag(FILE *f,const QString &tag,const QHostAddress &addr,
			   int indent) const
{
  WriteTag(f,tag,addr.toString(),indent);
}


void LiveWireMap::WriteTag(FILE *f,const QString &tag,unsigned value,int indent)
  const
{
  for(int i=0;i<indent;i++) {
    fprintf(f," ");
  }
  fprintf(f,"<%s>%u</%s>\n",(const char *)tag.toUtf8(),
	  value,(const char *)tag.toUtf8());
}


QString LiveWireMap::EscapeXml(const QString &str) const
{
  QString ret=str;
  ret.replace("&","&amp;");
  ret.replace("<","&lt;");
  ret.replace(">","&gt;");
  ret.replace("'","&apos;");
  ret.replace("\"","&quot;");

  return ret;
}
