// lwpathd.cpp
//
// lwpathd(8) routing daemon
//
//   (C) Copyright 2015 Fred Gleason <fredg@paravelsystems.com>
//
//   All Rights Reserved.
//

#include <signal.h>
#include <stdio.h>
#include <syslog.h>
#include <unistd.h>

#include <QApplication>
#include <QCoreApplication>

#include <sy/sycmdswitch.h>

#include "config.h"
#include "lwpathd.h"
#include "matrixfactory.h"
#include "protocolfactory.h"

void SignalHandler(int signo)
{
  switch(signo) {
  case SIGINT:
  case SIGTERM:
    unlink(LWPATHD_PID_FILE);
    exit(0);
  }
}


MainObject::MainObject(QObject *parent)
  : QObject(parent)
{
  debug=false;
  FILE *f=NULL;

  SyCmdSwitch *cmd=
    new SyCmdSwitch(qApp->argc(),qApp->argv(),"lwpathd",VERSION,LWPATHD_USAGE);
  for(unsigned i=0;i<cmd->keys();i++) {
    if(cmd->key(i)=="-d") {
      debug=true;
      cmd->setProcessed(i,true);
    }
    if(!cmd->processed(i)) {
      fprintf(stderr,"lwpathd: unknown option\n");
      exit(256);
    }
  }

  //
  // Open the Syslog
  //
  if(debug) {
    openlog("lwpathd",LOG_PERROR,LOG_DAEMON);
  }
  else {
    openlog("lwpathd",0,LOG_DAEMON);
  }

  //
  // Open Configuration File
  //
  path_config=new SyProfile();
  if(!path_config->setSource(LWPATH_CONF_FILE)) {
    fprintf(stderr,"lwpathd: unable to open configuration file\n");
    exit(256);
  }

  //
  // Load Resources
  //
  LoadMatrices();
  LoadProtocols();

  //
  // Detach
  //
  if(!debug) {
    daemon(0,0);
  }

  //
  // Install Signal Handlers
  //
  ::signal(SIGTERM,SignalHandler);
  ::signal(SIGINT,SignalHandler);

  //
  // Create PID File
  //
  if((f=fopen(LWPATHD_PID_FILE,"w"))!=NULL) {
    fprintf(f,"%u",getpid());
    fclose(f);
  }
}


void MainObject::startupCompleteData()
{
  bool all_started=true;
  for(unsigned i=0;i<path_matrices.size();i++) {
    all_started=all_started&&path_matrices[i]->isReady();
  }
  if(all_started) {
    for(unsigned i=0;i<path_protocols.size();i++) {
      path_protocols[i]->setReady(true);
    }
  }
}


void MainObject::LoadMatrices()
{
  int n=0;
  unsigned mtxnum=0;
  bool ok;
  QString str;

  str=path_config->
    stringValue("Matrices",QString().sprintf("Type%d",n+1),"",&ok);
  while(ok) {
    ok=false;
    if(str.toLower()=="livewire") {
      path_matrices.
	push_back(MatrixFactory(Matrix::TypeLiveWire,mtxnum,path_config,this));
      path_matrices.back()->
	setMatrixName(path_config->
		      stringValue("Matrices",QString().sprintf("Name%d",n+1),
				  Matrix::typeString(Matrix::TypeLiveWire)));
      connect(path_matrices.back(),SIGNAL(startupComplete()),
	      this,SLOT(startupCompleteData()));
      ok=true;
    }
    if(ok) {
      mtxnum+=path_matrices.back()->matrixQuantity();
    }
    else {
      fprintf(stderr,"lwpathd: unrecognized matrix type \"%s\"\n",
	      (const char *)str.toUtf8());
      exit(256);
    }
    n++;
    str=path_config->
      stringValue("Matrices",QString().sprintf("Type%d",n+1),"",&ok);
  }
}


void MainObject::LoadProtocols()
{
  int n=0;
  bool ok;
  QString str;

  str=path_config->
    stringValue("Protocols",QString().sprintf("Type%d",n+1),"",&ok);
  while(ok) {
    ok=false;
    for(int i=0;i<Protocol::TypeLast;i++) {
      if(str.toLower()==Protocol::keyword((Protocol::Type)i)) {
	path_protocols.
	  push_back(ProtocolFactory((Protocol::Type)i,&path_matrices,
				    path_config,this));
	ok=true;
      }
    }
    if(!ok) {
      fprintf(stderr,"lwpathd: unrecognized protocol type \"%s\"\n",
	      (const char *)str.toUtf8());
      exit(256);
    }
    n++;
    str=path_config->
      stringValue("Protocols",QString().sprintf("Type%d",n+1),"",&ok);
  }
}


int main(int argc,char *argv[])
{
  QCoreApplication a(argc,argv);
  new MainObject();
  return a.exec();
}
