// matrix_livewire.cpp
//
// Matrix driver for LiveWire.
//
//   (C) Copyright 2015 Fred Gleason <fredg@paravelsystems.com>
//
//   This program is free software; you can redistribute it and/or modify
//   it under the terms of the GNU General Public License version 2 as
//   published by the Free Software Foundation.
//
//   This program is distributed in the hope that it will be useful,
//   but WITHOUT ANY WARRANTY; without even the implied warranty of
//   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//   GNU General Public License for more details.
//
//   You should have received a copy of the GNU General Public
//   License along with this program; if not, write to the Free Software
//   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
//

#include <stdio.h>
#include <syslog.h>
#include <unistd.h>

#include <sy/syconfig.h>
#include <sy/symcastsocket.h>

#include "matrix_livewire.h"

MatrixLiveWire::MatrixLiveWire(int mid,SyProfile *config,QObject *parent)
  : Matrix(Matrix::TypeLiveWire,mid,config,parent)
{
  lw_startup_node=0;
  lw_load_complete=false;
  lw_disable_watchdog=config->boolValue("Matrices",
		      QString().sprintf("DisableWatchdog%d",mid+1),false);

  lw_map=new LiveWireMap();

  lw_gpi_reset_mapper=new QSignalMapper(this);
  connect(lw_gpi_reset_mapper,SIGNAL(mapped(int)),this,SLOT(gpiResetData(int)));

  lw_gpo_reset_mapper=new QSignalMapper(this);
  connect(lw_gpo_reset_mapper,SIGNAL(mapped(int)),this,SLOT(gpoResetData(int)));
  lw_node_connection_timeout=
    config->intValue("Matrices",QString().sprintf("NodeTimeout%d",mid+1),5000);

  LoadMap(config->stringValue("Matrices",QString().sprintf("MapFile%d",mid+1),
			      QString(SWITCHYARD_CACHE_DIR)+"/"+
			      LIVEWIRE_DEFAULT_MAP_FILE));
  //  lw_map->save("");
  LoadNodes();
}


MatrixLiveWire::~MatrixLiveWire()
{
  delete lw_gpo_reset_mapper;
  delete lw_gpi_reset_mapper;
}


unsigned MatrixLiveWire::matrixQuantity() const
{
  return 2;
}


QString MatrixLiveWire::matrixName(int mid) const
{
  if(mid==matrixId()) {  // Audio Matrix
    return Matrix::matrixName(mid)+" Audio";
  }
  else {                 // GPIO Matrix
    return Matrix::matrixName(mid)+" GPIO";
  }
}


bool MatrixLiveWire::isReady()
{
  return lw_load_complete;
}


unsigned MatrixLiveWire::nodeQuantity(int mid) const
{
  return lw_map->nodeQuantity();
}


bool MatrixLiveWire::nodeOnline(int mid,unsigned node) const
{
  return lw_map->node(node)->isOnline();
}


QString MatrixLiveWire::nodeHostName(int mid,unsigned node) const
{
  return lw_map->node(node)->hostName();
}


QHostAddress MatrixLiveWire::nodeHostAddress(int mid,unsigned node) const
{
  return lw_map->node(node)->hostAddress();
}


QString MatrixLiveWire::nodeDeviceName(int mid,unsigned node) const
{
  return lw_map->node(node)->deviceName();
}


unsigned MatrixLiveWire::gpioCodeSize(int mid) const
{
  return SWITCHYARD_GPIO_BUNDLE_SIZE;
}


unsigned MatrixLiveWire::inputQuantity(int mid) const
{
  if(mid==matrixId()) {  // Audio Matrix
    return lw_map->inputQuantity();
  }
  else {                 // GPIO Matrix
    return lw_map->gpiQuantity();
  }
}


QString MatrixLiveWire::inputName(int mid,unsigned input) const
{
  QString ret=Matrix::inputName(mid,input);

  if(mid==matrixId()) {  // Audio Matrix
    if(input<lw_map->inputQuantity()) {
      ret=lw_map->input(input)->name();
    }
  }
  else {                 // GPIO Matrix
    if(input<lw_map->gpiQuantity()) {
      ret=QString().sprintf("GPI %u",lw_map->gpi(input)->slotNumber()+1);
    }
  }

  return ret;
}


unsigned MatrixLiveWire::inputSlotNumber(int mid,unsigned input) const
{
  unsigned ret=Matrix::inputSlotNumber(mid,input);

  if(mid==matrixId()) {  // Audio Matrix
    if(input<lw_map->inputQuantity()) {
      ret=lw_map->input(input)->slotNumber();
    }
  }
  else {                 // GPIO Matrix
    if(input<lw_map->gpiQuantity()) {
      ret=lw_map->gpi(input)->slotNumber();
    }
  }

  return ret;
}


QString MatrixLiveWire::inputHostName(int mid,unsigned input) const
{
  QString ret=Matrix::inputHostName(mid,input);

  if(mid==matrixId()) {  // Audio Matrix
    if(input<lw_map->inputQuantity()) {
      ret=lw_map->input(input)->hostName();
    }
  }
  else {                 // GPIO Matrix
    if(input<lw_map->gpiQuantity()) {
      ret=lw_map->gpi(input)->hostName();
    }
  }

  return ret;
}


QHostAddress MatrixLiveWire::inputHostAddress(int mid,unsigned input) const
{
  QHostAddress ret=Matrix::inputHostAddress(mid,input);

  if(mid==matrixId()) {  // Audio Matrix
    if(input<lw_map->inputQuantity()) {
      ret=lw_map->input(input)->hostAddress();
    }
  }
  else {                 // GPIO Matrix
    if(input<lw_map->gpiQuantity()) {
      ret=lw_map->gpi(input)->hostAddress();
    }
  }

  return ret;
}


QHostAddress MatrixLiveWire::inputSourceAddress(int mid,unsigned input) const
{
  QHostAddress ret=Matrix::inputSourceAddress(mid,input);

  if(mid==matrixId()) {  // Audio Matrix
    if(input<lw_map->inputQuantity()) {
      ret=lw_map->input(input)->streamAddress();
    }
  }
  else {                 // GPIO Matrix
    if(input<lw_map->gpiQuantity()) {
      ret=lw_map->gpi(input)->hostAddress();
    }
  }

  return ret;
}


int MatrixLiveWire::inputSourceSlot(int mid,unsigned input) const
{
  int ret=Matrix::inputSourceSlot(mid,input);

  if(mid!=matrixId()) {  // Only GPIO sources have source slots
    if(input<lw_map->gpiQuantity()) {
      ret=lw_map->gpi(input)->slotNumber();
    }
  }

  return ret;
}


unsigned MatrixLiveWire::gpiQuantity(int mid) const
{
  if(mid!=matrixId()) {  // GPIO Matrix
    return lw_map->gpiQuantity();
  }
  return 0;
}


QString MatrixLiveWire::gpiCode(int mid,unsigned input) const
{
  if(mid!=matrixId()) {  // GPIO Matrix
    if(input<lw_map->gpiQuantity()) {
      return lw_map->gpi(input)->bundle()->code();
    }
  }
  return QString("hhhhh");
}


void MatrixLiveWire::setGpiCode(int mid,unsigned input,int msecs,
				const QString &code)
{
  SyLwrpClient *node=NULL;

  if(mid!=matrixId()) {  // GPIO Matrix
    if(input<lw_map->gpiQuantity()) {
      if(lw_lwrps.count(lw_map->gpi(input)->hostAddress().toIPv4Address())>0) {
	node=lw_lwrps[lw_map->gpi(input)->hostAddress().toIPv4Address()];
	if(msecs>0) {
	  lw_map->gpi(input)->
	    setSavedCode(node->gpiBundle(lw_map->gpi(input)->slotNumber())->code());
	  lw_map->gpi(input)->timer()->start(msecs);
	}
	node->setGpiCode(lw_map->gpi(input)->slotNumber(),code);
      }
    }
  }
}


unsigned MatrixLiveWire::outputQuantity(int mid) const
{
  if(mid==matrixId()) {  // Audio Matrix
    return lw_map->outputQuantity();
  }
  else {                 // GPIO Matrix
    return lw_map->gpoQuantity();
  }
}


QString MatrixLiveWire::outputName(int mid,unsigned output) const
{
  QString ret=Matrix::outputName(mid,output);

  if(mid==matrixId()) {  // Audio Matrix
    if(output<lw_map->outputQuantity()) {
      ret=lw_map->output(output)->name();
      if(ret.isEmpty()) {
	ret=QString().sprintf("DST %u",lw_map->output(output)->slotNumber()+1);
      }
    }
  }
  else {                 // GPIO Matrix
    if(output<lw_map->gpoQuantity()) {
      ret=lw_map->gpo(output)->name();
      if(ret.isEmpty()) {
	ret=QString().sprintf("GPO %u",lw_map->gpo(output)->slotNumber()+1);
      }
    }
  }

  return ret;
}


unsigned MatrixLiveWire::outputSlotNumber(int mid,unsigned output) const
{
  unsigned ret=Matrix::outputSlotNumber(mid,output);

  if(mid==matrixId()) {  // Audio Matrix
    if(output<lw_map->outputQuantity()) {
      ret=lw_map->output(output)->slotNumber();
    }
  }
  else {                 // GPIO Matrix
    if(output<lw_map->gpoQuantity()) {
      ret=lw_map->gpo(output)->slotNumber();
    }
  }

  return ret;
}


QString MatrixLiveWire::outputHostName(int mid,unsigned output) const
{
  QString ret=Matrix::outputHostName(mid,output);

  if(mid==matrixId()) {  // Audio Matrix
    if(output<lw_map->outputQuantity()) {
      ret=lw_map->output(output)->hostName();
    }
  }
  else {                 // GPIO Matrix
    if(output<lw_map->gpoQuantity()) {
      ret=lw_map->gpo(output)->hostName();
    }
  }

  return ret;
}


QHostAddress MatrixLiveWire::outputHostAddress(int mid,unsigned output) const
{
  QHostAddress ret=Matrix::outputHostAddress(mid,output);

  if(mid==matrixId()) {  // Audio Matrix
    if(output<lw_map->outputQuantity()) {
      ret=lw_map->output(output)->hostAddress();
    }
  }
  else {                 // GPIO Matrix
    if(output<lw_map->gpoQuantity()) {
      ret=lw_map->gpo(output)->hostAddress();
    }
  }

  return ret;
}


unsigned MatrixLiveWire::outputCrosspoint(int mid,unsigned output)
{
  LiveWireAudioport *in=NULL;
  LiveWireAudioport *out=NULL;
  LiveWireGpio *gpi=NULL;
  LiveWireGpio *gpo=NULL;
  unsigned ret=0;

  if(mid==matrixId()) {  // Audio Matrix
    if(output<=lw_map->outputQuantity()) {
      out=lw_map->output(output-1);
      if((in=GetInput(out->streamAddress()))!=NULL) {
	ret=in->number();
      }
    }
  }
  else {                 // GPIO Matrix
    if(output<=lw_map->outputQuantity()) {
      gpo=lw_map->gpo(output-1);
      if((gpi=GetGpi(gpo->sourceAddress().toIPv4Address(),
		     gpo->sourceSlot()))!=NULL) {
	ret=gpi->number();
      }
    }
  }

  return ret;
}


void MatrixLiveWire::setOutputCrosspoint(int mid,unsigned output,unsigned input)
{
  LiveWireAudioport *in=NULL;
  LiveWireAudioport *out=NULL;
  LiveWireGpio *gpi=NULL;
  LiveWireGpio *gpo=NULL;
  uint32_t nodeaddr=0;

  if(mid==matrixId()) {  // Audio Matrix
    if((input<lw_map->inputQuantity())&&(output<lw_map->outputQuantity())) {
      in=lw_map->input(input);
      out=lw_map->output(output);
      nodeaddr=out->hostAddress().toIPv4Address();
      if(lw_lwrps.count(nodeaddr)>0) {
	lw_lwrps[nodeaddr]->
	  setDstAddress(out->slotNumber(),in->streamAddress());
      }
    }
  }
  else {                 // GPIO Matrix
    if((input<lw_map->gpiQuantity())&&(output<lw_map->gpoQuantity())) {
      gpi=lw_map->gpi(input);
      gpo=lw_map->gpo(output);
      nodeaddr=gpo->hostAddress().toIPv4Address();
      lw_lwrps[nodeaddr]->
	setGpoSourceAddress(gpo->slotNumber(),gpi->hostAddress(),
			    gpi->slotNumber());
    }
  }
}


unsigned MatrixLiveWire::gpoQuantity(int mid) const
{
  if(mid!=matrixId()) {  // GPIO Matrix
    return lw_map->gpoQuantity();
  }
  return 0;
}


QString MatrixLiveWire::gpoCode(int mid,unsigned output) const
{
  if(mid!=matrixId()) {  // GPIO Matrix
    if(output<lw_map->gpoQuantity()) {
      return lw_map->gpo(output)->bundle()->code();
    }
  }
  return QString("hhhhh");
}


void MatrixLiveWire::setGpoCode(int mid,unsigned output,int msecs,
				const QString &code)
{
  SyLwrpClient *node=NULL;

  if(mid!=matrixId()) {  // GPIO Matrix
    if(output<lw_map->gpoQuantity()) {
      if(lw_lwrps.count(lw_map->gpo(output)->hostAddress().toIPv4Address())>0) {
	node=lw_lwrps[lw_map->gpo(output)->hostAddress().toIPv4Address()];
	if(msecs>0) {
	  lw_map->gpo(output)->
	    setSavedCode(node->gpo(lw_map->gpo(output)->slotNumber())->bundle()->
			 code());
	  lw_map->gpo(output)->timer()->start(msecs);
	}
	node->setGpoCode(lw_map->gpo(output)->slotNumber(),code);
      }
    }
  }
}


void MatrixLiveWire::connectedData(unsigned addr,bool state)
{
  LiveWireNode *node=NULL;
  LiveWireAudioport *in=NULL;
  LiveWireAudioport *out=NULL;
  LiveWireGpio *gpio=NULL;
  
  if((node=GetNode(QHostAddress(addr)))!=NULL) {
    if(state) {
      if(node->connected()) { 
	syslog(LOG_WARNING,"connection to node at %s restored",
	       (const char *)QHostAddress(addr).toString().toUtf8());
      }
    }
    else {
	syslog(LOG_WARNING,"connection to node at %s failed",
	       (const char *)QHostAddress(addr).toString().toUtf8());
    }
    node->makeConnected();
  }

  if(state) {
    //
    // Populate Node
    //
    node->setHostName(lw_lwrps[addr]->hostName());
    if(!node->isOnline()) {
      node->setOnline(true);
      emit nodeStatusAlarm(matrixId(),node->number(),true);
    }

    //
    // Populate Inputs
    //
    for(unsigned i=0;i<lw_lwrps[addr]->srcSlots();i++) {
      if((in=GetInput(addr,i))!=NULL) {
	in->setStreamAddress(lw_lwrps[addr]->srcAddress(i));
	lw_input_stream_index[in->streamAddress().toIPv4Address()]=in;
	in->setName(lw_lwrps[addr]->srcName(i));
	in->setHostName(lw_lwrps[addr]->hostName());
      }
    }

    //
    // Populate Outputs
    //
    for(unsigned i=0;i<lw_lwrps[addr]->dstSlots();i++) {
      if((out=GetOutput(addr,i))!=NULL) {
	out->setStreamAddress(lw_lwrps[addr]->dstAddress(i));
	out->setName(lw_lwrps[addr]->dstName(i));
	out->setHostName(lw_lwrps[addr]->hostName());
      }
    }

    //
    // Populate GPIs
    //
    for(unsigned i=0;i<lw_lwrps[addr]->gpis();i++) {
      if((gpio=GetGpi(addr,i))!=NULL) {
	gpio->bundle()->setCode(lw_lwrps[addr]->gpiBundle(i)->code());
	gpio->setHostName(lw_lwrps[addr]->hostName());
      }
    }

    //
    // Populate GPOs
    //
    for(unsigned i=0;i<lw_lwrps[addr]->gpos();i++) {
      if((gpio=GetGpo(addr,i))!=NULL) {
	gpio->bundle()->setCode(lw_lwrps[addr]->gpo(i)->bundle()->code());
	gpio->setName(lw_lwrps[addr]->gpo(i)->name());
	gpio->setHostName(lw_lwrps[addr]->hostName());
	gpio->setSourceAddress(lw_lwrps[addr]->gpo(i)->sourceAddress(),
			       lw_lwrps[addr]->gpo(i)->sourceSlot());
      }
    }

    //
    // Configure Audio Monitors
    //
    /*
     * FIXME: This throws errors on Mix Engines
     *
    for(unsigned i=0;i<lw_lwrps[addr]->srcSlots();i++) {
      lw_lwrps[addr]->setClipMonitor(i,SyLwrpClient::InputMeter,
				     10*clipThreshold(),clipTimeout());
      lw_lwrps[addr]->setSilenceMonitor(i,SyLwrpClient::InputMeter,
				     10*silenceThreshold(),silenceTimeout());
    }
    for(unsigned i=0;i<lw_lwrps[addr]->dstSlots();i++) {
      lw_lwrps[addr]->setClipMonitor(i,SyLwrpClient::OutputMeter,
				     10*clipThreshold(),clipTimeout());
      lw_lwrps[addr]->setSilenceMonitor(i,SyLwrpClient::OutputMeter,
					10*silenceThreshold(),silenceTimeout());
    }
    */
  }
  else {
    if(node->isOnline()) {
      node->setOnline(false);
      emit nodeStatusAlarm(matrixId(),node->number(),false);
    }
  }
  if((!lw_load_complete)&&(!StartNextNode())) {
    lw_load_complete=true;
    syslog(LOG_DEBUG,"startup complete");
    emit startupComplete();
  }
}


void MatrixLiveWire::connectionErrorData(unsigned addr,
					 QAbstractSocket::SocketError err)
{
  syslog(LOG_WARNING,"node at %s returned network error %d [%s]",
	 (const char *)QHostAddress(addr).toString().toUtf8(),err,
	 (const char *)SyMcastSocket::socketErrorText(err).toUtf8());
  connectedData(addr,false);
}


void MatrixLiveWire::sourceChangedData(unsigned addr,int slotnum,
				       const SyNode &node,const SySource &src)
{
  LiveWireAudioport *in=NULL;

  if((in=GetInput(addr,slotnum))!=NULL) {
    in->setName(src.name());
    in->setHostName(node.hostName());
    in->setStreamAddress(src.streamAddress());
    lw_input_stream_index[src.streamAddress().toIPv4Address()]=in;
    in->setName(src.name());
  }
}


void MatrixLiveWire::audioClipAlarmData(unsigned addr,
					SyLwrpClient::MeterType type,
					unsigned slotnum,int chan,bool state)
{
  LiveWireAudioport *port=NULL;

  switch(type) {
  case SyLwrpClient::InputMeter:
    if((port=GetInput(addr,slotnum))!=NULL) {
      emit inputClipAlarm(matrixId(),port->number(),state);
    }
    break;

  case SyLwrpClient::OutputMeter:
    if((port=GetOutput(addr,slotnum))!=NULL) {
      emit outputClipAlarm(matrixId(),port->number(),state);
    }
    break;

  case SyLwrpClient::LastTypeMeter:
    break;

  }
}


void MatrixLiveWire::audioSilenceAlarmData(unsigned addr,
					   SyLwrpClient::MeterType type,
					   unsigned slotnum,int chan,bool state)
{
  LiveWireAudioport *port=NULL;

  switch(type) {
  case SyLwrpClient::InputMeter:
    if((port=GetInput(addr,slotnum))!=NULL) {
      emit inputSilenceAlarm(matrixId(),port->number(),state);
    }
    break;

  case SyLwrpClient::OutputMeter:
    if((port=GetOutput(addr,slotnum))!=NULL) {
      emit outputSilenceAlarm(matrixId(),port->number(),state);
    }
    break;

  case SyLwrpClient::LastTypeMeter:
    break;
  }
}


void MatrixLiveWire::destinationChangedData(unsigned addr,int slotnum,
					    const SyNode &node,
					    const SyDestination &dst)
{
  LiveWireAudioport *in=NULL;
  LiveWireAudioport *out=NULL;

  if((out=GetOutput(addr,slotnum))!=NULL) {
    if(dst.streamAddress()!=out->streamAddress()) {
      if((in=GetInput(dst.streamAddress()))!=NULL) {
	emit outputCrosspointChanged(matrixId(),out->number(),in->number());
      }
      else {
	emit outputCrosspointChanged(matrixId(),out->number(),0);
      }
    }
    out->setStreamAddress(dst.streamAddress());
    out->setName(dst.name());
  }
}


void MatrixLiveWire::gpiChangedData(unsigned id,int slotnum,const SyNode &node,
				const SyGpioBundle &bundle)
{
  LiveWireGpio *gpio=NULL;

  if((gpio=GetGpi(node.hostAddress().toIPv4Address(),slotnum))!=NULL) {
    if(gpio->bundle()->code()!=bundle.code()) {
      gpio->bundle()->setCode(bundle.code());
      emit gpiChanged(matrixId()+1,gpio->number(),bundle.code());
    }
  }
}


void MatrixLiveWire::gpoChangedData(unsigned id,int slotnum,const SyNode &node,
				const SyGpo &gpo)
{
  LiveWireGpio *gpio_out=NULL;
  LiveWireGpio *gpio_in=NULL;

  if((gpio_out=GetGpo(node.hostAddress().toIPv4Address(),slotnum))!=NULL) {
    if(gpio_out->bundle()->code()!=gpo.bundle()->code()) {
      gpio_out->bundle()->setCode(gpo.bundle()->code());
      emit gpoChanged(matrixId()+1,gpio_out->number(),gpo.bundle()->code());
    }
    if((gpio_out->sourceAddress()!=gpo.sourceAddress())||
       (gpio_out->sourceSlot()!=gpo.sourceSlot())) {
      if((gpio_in=GetGpi(gpo.sourceAddress().toIPv4Address(),
			 gpo.sourceSlot()))!=NULL) {
	gpio_out->setSourceAddress(gpo.sourceAddress(),gpo.sourceSlot());
	emit outputCrosspointChanged(matrixId()+1,
				     gpio_out->number(),gpio_in->number());
      }
      else {
	emit outputCrosspointChanged(matrixId()+1,gpio_out->number(),0);
      }
    }
  }
}


void MatrixLiveWire::gpiResetData(int input)
{
  if(lw_lwrps.count(lw_map->gpi(input)->hostAddress().toIPv4Address())>0) {
    lw_lwrps[lw_map->gpi(input)->hostAddress().toIPv4Address()]->
      setGpiCode(lw_map->gpi(input)->slotNumber(),
		 lw_map->gpi(input)->savedCode());
    lw_map->gpi(input)->clearSavedCode();
  }
}


void MatrixLiveWire::gpoResetData(int output)
{
  if(lw_lwrps.count(lw_map->gpo(output)->hostAddress().toIPv4Address())>0) {
    lw_lwrps[lw_map->gpo(output)->hostAddress().toIPv4Address()]->
      setGpoCode(lw_map->gpo(output)->slotNumber(),
		 lw_map->gpo(output)->savedCode());
    lw_map->gpo(output)->clearSavedCode();
  }
}


void MatrixLiveWire::LoadNodes()
{
  syslog(LOG_DEBUG,"startup begins");
  lw_load_complete=false;
  for(unsigned i=0;i<lw_map->nodeQuantity();i++) {
    if(!lw_map->node(i)->hostAddress().isNull()) {
      uint32_t addr=lw_map->node(i)->hostAddress().toIPv4Address();
      lw_lwrps[addr]=new SyLwrpClient(addr,this);
      lw_lwrps[addr]->setTimeoutInterval(lw_node_connection_timeout);
      connect(lw_lwrps[addr],SIGNAL(connected(unsigned,bool)),
	      this,SLOT(connectedData(unsigned,bool)));
      connect(lw_lwrps[addr],
	      SIGNAL(connectionError(unsigned,QAbstractSocket::SocketError)),
	      this,
	      SLOT(connectionErrorData(unsigned,QAbstractSocket::SocketError)));
      connect(lw_lwrps[addr],
	      SIGNAL(sourceChanged(unsigned,int,const SyNode &,
				   const SySource &)),
	      this,SLOT(sourceChangedData(unsigned,int,const SyNode &,
					  const SySource &)));
      connect(lw_lwrps[addr],
	      SIGNAL(destinationChanged(unsigned,int,const SyNode &,
					const SyDestination &)),
	      this,SLOT(destinationChangedData(unsigned,int,const SyNode &,
					       const SyDestination &)));
      connect(lw_lwrps[addr],
	      SIGNAL(audioSilenceAlarm(unsigned,SyLwrpClient::MeterType,
				       unsigned,int,bool)),
	      this,SLOT(audioSilenceAlarmData(unsigned,SyLwrpClient::MeterType,
					      unsigned,int,bool)));
      connect(lw_lwrps[addr],
	      SIGNAL(gpiChanged(unsigned,int,const SyNode &,
				const SyGpioBundle &)),
	      this,SLOT(gpiChangedData(unsigned,int,const SyNode &,
				       const SyGpioBundle &)));
      connect(lw_lwrps[addr],
	      SIGNAL(gpoChanged(unsigned,int,const SyNode &,const SyGpo &)),
	      this,SLOT(gpoChangedData(unsigned,int,const SyNode &,
				       const SyGpo &)));
      lw_startup_addresses.push_back(addr);
    }
  }
  lw_startup_node=0;
  StartNextNode();
}


bool MatrixLiveWire::StartNextNode()
{
  LiveWireNode *node;

  if(lw_startup_node<lw_map->nodeQuantity()) {
    node=lw_map->node(lw_startup_node++);
    syslog(LOG_DEBUG,"starting node: %s",
	   (const char *)node->hostAddress().toString().toUtf8());
    lw_lwrps[node->hostAddress().toIPv4Address()]->
      connectToHost(node->hostAddress(),SWITCHYARD_LWRP_PORT,
		    node->password(),true);
    return true;
  }

  //
  // Index any "OFF" sources
  //
  for(unsigned i=0;i<lw_map->inputQuantity();i++) {
    LiveWireAudioport *in=lw_map->input(i);
    if(in->hostAddress().isNull()) {
      lw_input_stream_index[0]=in;
      lw_input_stream_index[QHostAddress("239.192.0.0").toIPv4Address()]=in;
    }
  }
  return false;
}


LiveWireNode *MatrixLiveWire::GetNode(const QHostAddress &addr) const
{
  for(unsigned i=0;i<lw_map->nodeQuantity();i++) {
    if(lw_map->node(i)->hostAddress()==addr) {
      return lw_map->node(i);
    }
  }
  return NULL;
}


LiveWireAudioport *MatrixLiveWire::GetInput(const QHostAddress &s_addr)
{
  if(lw_input_stream_index.count(s_addr.toIPv4Address())==0) {
    return NULL;
  }
  return lw_input_stream_index[s_addr.toIPv4Address()];
}


LiveWireAudioport *MatrixLiveWire::GetInput(uint32_t hostaddr,
					    unsigned slotnum) const
{
  return lw_map->input(QHostAddress(hostaddr),slotnum);
}


LiveWireAudioport *MatrixLiveWire::GetOutput(uint32_t hostaddr,unsigned slotnum) const
{
  return lw_map->output(QHostAddress(hostaddr),slotnum);
}


LiveWireGpio *MatrixLiveWire::GetGpi(uint32_t hostaddr,unsigned slotnum) const
{
  return lw_map->gpi(QHostAddress(hostaddr),slotnum);
}


LiveWireGpio *MatrixLiveWire::GetGpo(uint32_t hostaddr,unsigned slotnum) const
{
  return lw_map->gpo(QHostAddress(hostaddr),slotnum);
}


void MatrixLiveWire::LoadMap(const QString &filename)
{
  QString error_msg;

  if(!lw_map->load(filename,&error_msg)) {
    fprintf(stderr,"lwpathd: %s of \"%s\"\n",
	    (const char *)error_msg.toUtf8(),
	    (const char *)filename.toUtf8());
    exit(256);
  }
  if(!lw_map->validate(&error_msg)) {
    fprintf(stderr,"lwpathd: map errors: %s\n",
	    (const char *)error_msg.toUtf8());
    exit(256);
  }
  for(unsigned i=0;i<lw_map->gpiQuantity();i++) {
    lw_gpi_reset_mapper->setMapping(lw_map->gpi(i)->timer(),i);
    connect(lw_map->gpi(i)->timer(),SIGNAL(timeout()),
	    lw_gpi_reset_mapper,SLOT(map()));
  }
  for(unsigned i=0;i<lw_map->gpoQuantity();i++) {
    lw_gpo_reset_mapper->setMapping(lw_map->gpo(i)->timer(),i);
    connect(lw_map->gpo(i)->timer(),SIGNAL(timeout()),
	    lw_gpo_reset_mapper,SLOT(map()));
  }
}
